package sun.awt.geom;

import java.awt.geom.PathIterator;
import java.util.Enumeration;
import java.util.Vector;

public abstract class Crossings
{
    public static final boolean debug = false;
    int limit;
    double[] yranges;
    double xlo;
    double ylo;
    double xhi;
    double yhi;
    private Vector tmp;
    
    public Crossings(double xlo, double ylo, double xhi, double yhi) {
        this.limit = 0;
        this.yranges = new double[10];
        this.tmp = new Vector();
        this.xlo = xlo;
        this.ylo = ylo;
        this.xhi = xhi;
        this.yhi = yhi;
    }
    
    public final double getXLo() {
        return this.xlo;
    }
    
    public final double getYLo() {
        return this.ylo;
    }
    
    public final double getXHi() {
        return this.xhi;
    }
    
    public final double getYHi() {
        return this.yhi;
    }
    
    public abstract void record(double p0, double p1, int p2);
    
    public void print() {
        System.out.println("Crossings [");
        System.out.println("  bounds = [" + this.ylo + ", " + this.yhi + "]");
        for (int i = 0; i < this.limit; i += 2) {
            System.out.println("  [" + this.yranges[i] + ", " + this.yranges[i + 1] + "]");
        }
        System.out.println("]");
    }
    
    public final boolean isEmpty() {
        return this.limit == 0;
    }
    
    public abstract boolean covers(double p0, double p1);
    
    public static Crossings findCrossings(Vector vector, double n, double n2, double n3, double n4) {
        Crossings.EvenOdd evenOdd = new EvenOdd(n, n2, n3, n4);
        Enumeration<Object> elements = vector.elements();
        while (elements.hasMoreElements()) {
            if (((Curve)elements.nextElement()).accumulateCrossings(evenOdd)) {
                return null;
            }
            continue;
        }
        return evenOdd;
    }
    
    public static Crossings findCrossings(PathIterator pathIterator, double n, double n2, double n3, double n4) {
        Crossings.EvenOdd evenOdd;
        if (pathIterator.getWindingRule() == 0) {
            evenOdd = new EvenOdd(n, n2, n3, n4);
        }
        else {
            evenOdd = new NonZero(n, n2, n3, n4);
        }
        double[] array = new double[23];
        double n5 = 0.0;
        double n6 = 0.0;
        double n7 = 0.0;
        double n8 = 0.0;
        while (!pathIterator.isDone()) {
            switch (pathIterator.currentSegment(array)) {
                case 0: {
                    if (n6 != n8 && evenOdd.accumulateLine(n7, n8, n5, n6)) {
                        return null;
                    }
                    n7 = (n5 = array[0]);
                    n8 = (n6 = array[1]);
                    break;
                }
                case 1: {
                    double n9 = array[0];
                    double n10 = array[1];
                    if (evenOdd.accumulateLine(n7, n8, n9, n10)) {
                        return null;
                    }
                    n7 = n9;
                    n8 = n10;
                    break;
                }
                case 2: {
                    double n11 = array[2];
                    double n12 = array[3];
                    if (evenOdd.accumulateQuad(n7, n8, array)) {
                        return null;
                    }
                    n7 = n11;
                    n8 = n12;
                    break;
                }
                case 3: {
                    double n13 = array[4];
                    double n14 = array[5];
                    if (evenOdd.accumulateCubic(n7, n8, array)) {
                        return null;
                    }
                    n7 = n13;
                    n8 = n14;
                    break;
                }
                case 4: {
                    if (n6 != n8 && evenOdd.accumulateLine(n7, n8, n5, n6)) {
                        return null;
                    }
                    n7 = n5;
                    n8 = n6;
                    break;
                }
            }
            pathIterator.next();
        }
        if (n6 != n8 && evenOdd.accumulateLine(n7, n8, n5, n6)) {
            return null;
        }
        return evenOdd;
    }
    
    public boolean accumulateLine(double n, double n2, double n3, double n4) {
        if (n2 <= n4) {
            return this.accumulateLine(n, n2, n3, n4, 1);
        }
        return this.accumulateLine(n3, n4, n, n2, -1);
    }
    
    public boolean accumulateLine(double n, double n2, double n3, double n4, int n5) {
        if (this.yhi <= n2 || this.ylo >= n4) {
            return false;
        }
        if (n >= this.xhi && n3 >= this.xhi) {
            return false;
        }
        if (n2 == n4) {
            return n >= this.xlo || n3 >= this.xlo;
        }
        double n6 = n3 - n;
        double n7 = n4 - n2;
        double n8;
        double ylo;
        if (n2 < this.ylo) {
            n8 = n + (this.ylo - n2) * n6 / n7;
            ylo = this.ylo;
        }
        else {
            n8 = n;
            ylo = n2;
        }
        double n9;
        double yhi;
        if (this.yhi < n4) {
            n9 = n + (this.yhi - n2) * n6 / n7;
            yhi = this.yhi;
        }
        else {
            n9 = n3;
            yhi = n4;
        }
        if (n8 >= this.xhi && n9 >= this.xhi) {
            return false;
        }
        if (n8 > this.xlo || n9 > this.xlo) {
            return true;
        }
        this.record(ylo, yhi, n5);
        return false;
    }
    
    public boolean accumulateQuad(double n, double n2, double[] array) {
        if (n2 < this.ylo && array[1] < this.ylo && array[3] < this.ylo) {
            return false;
        }
        if (n2 > this.yhi && array[1] > this.yhi && array[3] > this.yhi) {
            return false;
        }
        if (n > this.xhi && array[0] > this.xhi && array[2] > this.xhi) {
            return false;
        }
        if (n < this.xlo && array[0] < this.xlo && array[2] < this.xlo) {
            if (n2 < array[3]) {
                this.record(Math.max(n2, this.ylo), Math.min(array[3], this.yhi), 1);
            }
            else if (n2 > array[3]) {
                this.record(Math.max(array[3], this.ylo), Math.min(n2, this.yhi), -1);
            }
            return false;
        }
        else {
            Curve.insertQuad(this.tmp, n, n2, array);
            Enumeration<Object> elements = this.tmp.elements();
            while (elements.hasMoreElements()) {
                if (((Curve)elements.nextElement()).accumulateCrossings(this)) {
                    return true;
                }
                continue;
            }
            this.tmp.clear();
            return false;
        }
    }
    
    public boolean accumulateCubic(double n, double n2, double[] array) {
        if (n2 < this.ylo && array[1] < this.ylo && array[3] < this.ylo && array[5] < this.ylo) {
            return false;
        }
        if (n2 > this.yhi && array[1] > this.yhi && array[3] > this.yhi && array[5] > this.yhi) {
            return false;
        }
        if (n > this.xhi && array[0] > this.xhi && array[2] > this.xhi && array[4] > this.xhi) {
            return false;
        }
        if (n < this.xlo && array[0] < this.xlo && array[2] < this.xlo && array[4] < this.xlo) {
            if (n2 <= array[5]) {
                this.record(Math.max(n2, this.ylo), Math.min(array[5], this.yhi), 1);
            }
            else {
                this.record(Math.max(array[5], this.ylo), Math.min(n2, this.yhi), -1);
            }
            return false;
        }
        else {
            Curve.insertCubic(this.tmp, n, n2, array);
            Enumeration<Object> elements = this.tmp.elements();
            while (elements.hasMoreElements()) {
                if (((Curve)elements.nextElement()).accumulateCrossings(this)) {
                    return true;
                }
                continue;
            }
            this.tmp.clear();
            return false;
        }
    }
    
    public static final class EvenOdd extends Crossings
    {
        public EvenOdd(double n, double n2, double n3, double n4) {
            super(n, n2, n3, n4);
        }
        
        @Override
        public final boolean covers(double n, double n2) {
            return this.limit == 2 && this.yranges[0] <= n && this.yranges[1] >= n2;
        }
        
        @Override
        public void record(double n, double n2, int n3) {
            if (n >= n2) {
                return;
            }
            int i;
            for (i = 0; i < this.limit && n > this.yranges[i + 1]; i += 2) {}
            while (i < this.limit) {
                double n4 = this.yranges[i++];
                double n5 = this.yranges[i++];
                if (n2 < n4) {
                    this.yranges[i++] = n;
                    this.yranges[i++] = n2;
                    n = n4;
                    n2 = n5;
                }
                else {
                    if (n < n4) {
                        n = n4;
                    }
                    else {
                        n = (n = n4);
                    }
                    if (n2 < n5) {
                        n = n2;
                    }
                    else {
                        n = n5;
                        n5 = n2;
                    }
                    if (n == n) {
                        n2 = n5;
                    }
                    else {
                        if (n > n) {
                            n = (n = n);
                        }
                        if (n != n) {
                            this.yranges[i++] = n;
                            this.yranges[i++] = n;
                        }
                        n2 = n5;
                    }
                    if (n >= n2) {
                        break;
                    }
                    continue;
                }
            }
            if (i < i && i < this.limit) {
                System.arraycopy(this.yranges, i, this.yranges, i, this.limit - i);
            }
            int limit = i + (this.limit - i);
            if (n < n2) {
                if (limit >= this.yranges.length) {
                    double[] yranges = new double[limit + 10];
                    System.arraycopy(this.yranges, 0, yranges, 0, limit);
                    this.yranges = yranges;
                }
                this.yranges[limit++] = n;
                this.yranges[limit++] = n2;
            }
            this.limit = limit;
            return;
        }
    }
    
    public static final class NonZero extends Crossings
    {
        private int[] crosscounts;
        
        public NonZero(double n, double n2, double n3, double n4) {
            super(n, n2, n3, n4);
            this.crosscounts = new int[this.yranges.length / 2];
        }
        
        @Override
        public final boolean covers(double n, double n2) {
            int i = 0;
            while (i < this.limit) {
                double n3 = this.yranges[i++];
                double n4 = this.yranges[i++];
                if (n >= n4) {
                    continue;
                }
                if (n < n3) {
                    return false;
                }
                if (n2 <= n4) {
                    return true;
                }
                n = n4;
            }
            return n >= n2;
        }
        
        public void remove(int n) {
            this.limit -= 2;
            int n2 = this.limit - n;
            if (n2 > 0) {
                System.arraycopy(this.yranges, n + 2, this.yranges, n, n2);
                System.arraycopy(this.crosscounts, n / 2 + 1, this.crosscounts, n / 2, n2 / 2);
            }
        }
        
        public void insert(int n, double n2, double n3, int n4) {
            int n5 = this.limit - n;
            double[] yranges = this.yranges;
            int[] crosscounts = this.crosscounts;
            if (this.limit >= this.yranges.length) {
                System.arraycopy(yranges, 0, this.yranges = new double[this.limit + 10], 0, n);
                System.arraycopy(crosscounts, 0, this.crosscounts = new int[(this.limit + 10) / 2], 0, n / 2);
            }
            if (n5 > 0) {
                System.arraycopy(yranges, n, this.yranges, n + 2, n5);
                System.arraycopy(crosscounts, n / 2, this.crosscounts, n / 2 + 1, n5 / 2);
            }
            this.yranges[n + 0] = n2;
            this.yranges[n + 1] = n3;
            this.crosscounts[n / 2] = n4;
            this.limit += 2;
        }
        
        @Override
        public void record(double n, double a, int n2) {
            if (n >= a) {
                return;
            }
            int n3;
            for (n3 = 0; n3 < this.limit && n > this.yranges[n3 + 1]; n3 += 2) {}
            if (n3 < this.limit) {
                int n4 = this.crosscounts[n3 / 2];
                double n5 = this.yranges[n3 + 0];
                double b = this.yranges[n3 + 1];
                if (b == n && n4 == n2) {
                    if (n3 + 2 == this.limit) {
                        this.yranges[n3 + 1] = a;
                        return;
                    }
                    this.remove(n3);
                    n = n5;
                    n4 = this.crosscounts[n3 / 2];
                    n5 = this.yranges[n3 + 0];
                    b = this.yranges[n3 + 1];
                }
                if (a < n5) {
                    this.insert(n3, n, a, n2);
                    return;
                }
                if (a == n5 && n4 == n2) {
                    this.yranges[n3] = n;
                    return;
                }
                if (n < n5) {
                    this.insert(n3, n, n5, n2);
                    n3 += 2;
                    n = n5;
                }
                else if (n5 < n) {
                    this.insert(n3, n5, n, n4);
                    n3 += 2;
                }
                int n6 = n4 + n2;
                double min = Math.min(a, b);
                if (n6 == 0) {
                    this.remove(n3);
                }
                else {
                    this.crosscounts[n3 / 2] = n6;
                    this.yranges[n3++] = n;
                    this.yranges[n3++] = min;
                }
                double n7 = n = min;
                if (n7 < b) {
                    this.insert(n3, n7, b, n4);
                }
            }
            if (n < a) {
                this.insert(n3, n, a, n2);
            }
            return;
        }
    }
}
